from enum import IntEnum

# Add new enums to __init__.py imports

class ConsoleColor(IntEnum):
    BLACK = 0
    DARK_RED = 1
    DARK_GREEN = 2
    DARK_YELLOW = 3
    DARK_BLUE = 4
    DARK_MAGENTA = 5
    DARK_CYAN = 6
    GRAY = 7

    DARK_GRAY = 10
    RED = 11
    GREEN = 12
    YELLOW = 13
    BLUE = 14
    MAGENTA = 15
    CYAN = 16
    WHITE = 17

class ConsoleStyle(IntEnum):
    BOLD = 1
    FAINT = 2
    ITALIC = 3
    UNDERLINE = 4
    STRIKE = 9



from nalpy.humanizer.inflections import vocabulary as _vocabulary
from nalpy.humanizer.numbers import count as _n_count
from nalpy.humanizer.numbers import enums as _n_enums

def pluralize(word: str, is_known_to_be_singular: bool = True) -> str:
    return _vocabulary.default().pluralize(word, is_known_to_be_singular)

def singularize(word: str, is_known_to_be_plural: bool = True) -> str:
    return _vocabulary.default().singularize(word, is_known_to_be_plural)

def quantisize(singular_word: str, amount: int, _number_format: _n_enums.NumberFormat = _n_enums.NumberFormat.NUMERALS) -> str:
    """Pluralizes the word with complex rules."""
    word: str
    if amount == 1:
        word = singular_word
    else:
        word = pluralize(singular_word)

    num: str = _n_count._format(amount, _number_format)
    if len(num) > 0:
        return f"{num} {word}"
    else:
        return word

def quick_quantisize(singular_word: str, amount: int, _number_format: _n_enums.NumberFormat = _n_enums.NumberFormat.NUMERALS) -> str:
    """Pluralizes the word by simply adding character ``s`` at the end of the word."""
    word = singular_word
    if amount != 1:
        word += "s"

    num: str = _n_count._format(amount, _number_format)
    if len(num) > 0:
        return f"{num} {word}"
    else:
        return word



from typing import Final, Iterable, NamedTuple, Self, final

from .. import Vector2Int, Rect


@final
class RectInt(NamedTuple):
    """
    A 2D Rectangle defined by X and Y position, width and height with integer precision.

    The functionality of this rect depends on the coordinate space used.
    When X increases to the right and Y increases upwards, position represents the bottom-left corner of the rectangle.
    When X increases to the right and Y increases downwards, position represents the top-left corner of the rectange.
    """

    x: int
    y: int
    w: int
    h: int

    @classmethod
    @property
    def zero(cls) -> Self:
        """Shorthand for ``math.RectInt(0, 0, 0, 0)``"""
        return _ZERO #  Returning single instance, because Rect is immutable

    #region Constructors
    @classmethod
    def from_pos(cls, position: Vector2Int, size: Vector2Int) -> Self:
        return cls(position.x, position.y, size.x, size.y)

    @classmethod
    def from_min_max(cls, xmin: int, ymin: int, xmax: int, ymax: int) -> Self:
        w: int = xmax - xmin
        h: int = ymax - ymin

        return cls(xmin, ymin, w, h)

    @classmethod
    def from_min_max_vectors(cls, min: Vector2Int, max: Vector2Int) -> Self:
        return cls.from_min_max(min.x, min.y, max.x, max.y)
    #endregion

    #region Operators

    # Provides __str__ also
    def __repr__(self) -> str:
        return f"RectInt({self.x}, {self.y}, {self.w}, {self.h})"
    #endregion

    #region Properties
    @property
    def min(self) -> Vector2Int:
        """Same as `Rect.position`"""
        return Vector2Int(self.xmin, self.ymin)
    @property
    def xmin(self) -> int:
        """Same as `Rect.x`"""
        return self.x
    @property
    def ymin(self) -> int:
        """Same as `Rect.y`"""
        return self.y

    @property
    def max(self) -> Vector2Int:
        return Vector2Int(self.xmax, self.ymax)
    @property
    def xmax(self) -> int:
        return self.x + self.w
    @property
    def ymax(self) -> int:
        return self.y + self.h

    @property
    def position(self) -> Vector2Int:
        return Vector2Int(self.x, self.y)
    @property
    def size(self) -> Vector2Int:
        return Vector2Int(self.w, self.h)

    @property
    def all_positions_within(self) -> Iterable[Vector2Int]:
        for y in range(self.ymin, self.ymax): # Return by row. (x first then y)
            for x in range(self.xmin, self.xmax):
                yield Vector2Int(x, y)
    #endregion

    #region Collisions
    def contains(self, point: Vector2Int) -> bool:
        return (point.x >= self.x) and (point.x < self.xmax) and (point.y >= self.y) and (point.y < self.ymax)

    def overlaps(self, other: Self) -> bool:
        return (
            other.xmax > self.x and
            other.x    < self.xmax and
            other.ymax > self.y and
            other.y    < self.ymax
        )
    #endregion

    #region Conversions
    def clamped_to_bounds(self, bounds: Self) -> Self:
        x: int = max(min(bounds.xmax, self.x), bounds.x)
        y: int = max(min(bounds.ymax, self.y), bounds.y)
        w: int = min(bounds.xmax - self.x, self.w)
        h: int = min(bounds.ymax - self.y, self.h)

        return RectInt(x, y, w, h)

    def to_rect(self) -> Rect:
        return Rect(float(self.x), float(self.y), float(self.w), float(self.h))

    def to_int_dict(self) -> dict[str, int]:
        return self._asdict()
    #endregion

_ZERO: Final[RectInt] = RectInt(0, 0, 0, 0)



"""
Provides C#-like Events.
"""

import typing as _typing


ParamsT = _typing.ParamSpec("ParamsT")

class Event(_typing.Generic[ParamsT]):
    """
    The event class on which other methods can subscribe using the `+=` operator and unsubscribe using `-=`. Can be typehinted.
    """
    def __init__(self, values: _typing.Iterable[_typing.Callable[ParamsT, _typing.Any]] = []) -> None:
        self.listeners = [v for v in values]

    def __iadd__(self, value: _typing.Callable[ParamsT, _typing.Any]) -> _typing.Self:
        if not callable(value):
            raise TypeError(f"Value of type {type(value)} is not callable!")

        if value in self.listeners:
            raise ValueError(f"Value {value} already in events!")

        self.listeners.append(value)

        return self

    def __isub__(self, value: _typing.Callable[ParamsT, _typing.Any])  -> _typing.Self:
        if not callable(value):
            raise TypeError(f"Value of type {type(value)} is not callable!")

        while value in self.listeners:
            self.listeners.remove(value)

        return self

    def __contains__(self, value: _typing.Callable[ParamsT, _typing.Any]) -> bool:
        return value in self.listeners

    def __len__(self) -> int:
        return len(self.listeners)

    def Invoke(self, *args: ParamsT.args, **kwargs: ParamsT.kwargs) -> None:
        for listener in self.listeners.copy(): # copying so that if someone removes themselves from the list, it still works
            listener(*args, **kwargs)



from typing import NamedTuple, final

from .. import RectInt, RectOffset

@final
class RectOffsetInt(NamedTuple):
    """
    Offsets for rectangles, borders, etc. with integer precision.

    Assumes that X increases to the right and Y increases downwards.
    """

    left: int
    right: int
    top: int
    bottom: int

    @property
    def horizontal(self) -> int:
        return self.left + self.right

    @property
    def vertical(self) -> int:
        return self.top + self.bottom

    def add(self, rect: RectInt) -> RectInt:
        """Add the border offsets to a rect."""

        return RectInt.from_min_max(
            rect.xmin + self.left,
            rect.ymin + self.top,
            rect.xmax - self.right,
            rect.ymax - self.bottom
        )

    def remove(self, rect: RectInt) -> RectInt:
        """Remove the border offsets from a rect."""

        return RectInt.from_min_max(
            rect.xmin - self.left,
            rect.ymin - self.top,
            rect.xmax + self.right,
            rect.ymax + self.bottom
        )

    def to_rect_offset(self) -> RectOffset:
        return RectOffset(float(self.left), float(self.right), float(self.top), float(self.bottom))

    def to_int_tuple(self) -> tuple[int, int, int, int]:
        left: int = round(self.left)
        right: int = round(self.right)
        top: int = round(self.top)
        bottom: int = round(self.bottom)
        return (left, right, top, bottom)

    def to_int_dict(self) -> dict[str, int]:
        return self._asdict()



from typing import ClassVar, Iterator, SupportsIndex
from .vector2 import Vector2

class Vector2Int:
    """An immutable two-dimensional vector with integer precision."""
    zero: ClassVar[Vector2Int]
    """Shorthand for ``Vector2Int(0, 0)``"""
    one: ClassVar[Vector2Int]
    """Shorthand for ``Vector2Int(1, 1)``"""
    up: ClassVar[Vector2Int]
    """A unit vector pointing up (vector j). Shorthand for ``Vector2Int(0, 1)``"""
    down: ClassVar[Vector2Int]
    """A unit vector pointing down. Shorthand for ``Vector2Int(0, -1)``"""
    left: ClassVar[Vector2Int]
    """A unit vector pointing left. Shorthand for ``Vector2Int(-1, 0)``"""
    right: ClassVar[Vector2Int]
    """A unit vector pointing right (vector i). Shorthand for ``Vector2Int(1, 0)``"""

    def __init__(self, x: int, y: int) -> None: ...

    def __getitem__(self, i: SupportsIndex) -> int: ...

    @property
    def x(self) -> int: ...

    @property
    def y(self) -> int: ...

    def __repr__(self) -> str: ...

    def __len__(self) -> int: ...

    def __add__(self, other: Vector2Int) -> Vector2Int: ...

    def __sub__(self, other: Vector2Int) -> Vector2Int: ...

    def __mul__(self, other: Vector2Int | int) -> Vector2Int: ...

    def __rmul__(self, other: Vector2Int | int) -> Vector2Int: ...

    def __truediv__(self, other: Vector2Int | int | float) -> Vector2: ...

    def __floordiv__(self, other: Vector2Int | int) -> Vector2Int: ...

    def __mod__(self, other: Vector2Int | int) -> Vector2Int: ...

    def __divmod__(self, other: Vector2Int | int) -> tuple[Vector2Int, Vector2Int]: ...

    def __neg__(self) -> Vector2Int: ...

    def __abs__(self) -> Vector2Int: ...

    def __eq__(self, other: Vector2Int) -> bool: ...

    def __iter__(self) -> Iterator[int]: ...

    def __hash__(self) -> int: ...

    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        ...

    @staticmethod
    def distance(a: Vector2Int, b: Vector2Int) -> float:
        """Returns the distance between a and b."""
        ...

    @staticmethod
    def ceil(v: Vector2) -> Vector2Int:
        """Create a new `Vector2Int` by ceiling the components of the given `Vector2`."""
        ...

    @staticmethod
    def floor(v: Vector2) -> Vector2Int:
        """Create a new `Vector2Int` by flooring the components of the given `Vector2`."""
        ...

    @staticmethod
    def round(v: Vector2) -> Vector2Int:
        """Create a new `Vector2Int` by rounding the components of the given `Vector2`."""
        ...

    @staticmethod
    def trunc(v: Vector2) -> Vector2Int:
        """Create a new `Vector2Int` by truncating the components of the given `Vector2`. Truncating is equivalent to `int(float_value)`."""
        ...

    @staticmethod
    def min(a: Vector2Int, b: Vector2Int) -> Vector2Int:
        """Create a new `Vector2Int` by selecting the smallest components of the two given `Vector2Int` instances."""
        ...

    @staticmethod
    def max(a: Vector2Int, b: Vector2Int) -> Vector2Int:
        """Create a new `Vector2Int` by selecting the largest components of the two given `Vector2Int` instances."""
        ...

    def to_vector2(self) -> Vector2:
        """`Vector2Int` to `Vector2` conversion."""
        ...

    def to_tuple(self) -> tuple[int, int]: ...
    def to_dict(self) -> dict[str, int]: ...



import sys
from nalpy.console_utils import ConsoleColor

ESCAPE = "\u001B"

def set_control(command: str):
    sys.stdout.write(command)

def set_csi(command: str):
    set_control(ESCAPE + "[" + command)

def set_seq(value: int, suffix: str):
    set_csi(f"{value}{suffix}")

def get_ansi_color(color: ConsoleColor, prefix_number: int) -> str:
    suffix = ""
    number = color.value
    if number >= 10:
        assert number < 20 # Just to make sure I haven't fucked something up in ConsoleColors
        suffix = ";1"
        number -= 10

    return f"{prefix_number}{number}{suffix}"

def set_sgr(command: str):
    set_csi(command + "m")

def erase_line(mode: int):
    set_seq(mode, "K")

def erase_display(mode: int):
    set_seq(mode, "J")



import unittest

from nalpy import math


class BasicMathFunctions(unittest.TestCase):
    def test_cbrt(self):
        self.assertAlmostEqual(math.cbrt(64), 4)
        self.assertAlmostEqual(math.cbrt(27), 3)
        self.assertAlmostEqual(math.cbrt(8), 2)
        self.assertAlmostEqual(math.cbrt(-8), -2)
        self.assertAlmostEqual(math.cbrt(0), 0)
        self.assertAlmostEqual(math.cbrt(1), 1)
        self.assertAlmostEqual(math.cbrt(-1), -1)

        self.assertAlmostEqual(math.cbrt(9), 2.08008382305190411453)
        self.assertAlmostEqual(math.cbrt(0.5), 0.79370052598409973738)
        self.assertAlmostEqual(math.cbrt(69), 4.10156592970234752185)
        self.assertAlmostEqual(math.cbrt(420), 7.48887238721850719787)

    def test_sign(self):
        self.assertEqual(math.sign(0), 0)
        self.assertEqual(math.sign(0.0), 0)
        self.assertEqual(math.sign(1), 1)
        self.assertEqual(math.sign(1.0), 1)
        self.assertEqual(math.sign(-1), -1)
        self.assertEqual(math.sign(-1.0), -1)

        self.assertEqual(math.sign(69), 1)
        self.assertEqual(math.sign(69.420), 1)
        self.assertEqual(math.sign(420.69), 1)

        self.assertEqual(math.sign(-69), -1)
        self.assertEqual(math.sign(-69.420), -1)
        self.assertEqual(math.sign(-420.69), -1)

    def test_is_positive_inf(self):
        self.assertEqual(math.is_positive_inf(float("inf")), True)
        self.assertEqual(math.is_positive_inf(float("-inf")), False)
        self.assertEqual(math.is_positive_inf(float("nan")), False)
        self.assertEqual(math.is_positive_inf(69), False)
        self.assertEqual(math.is_positive_inf(420), False)
        self.assertEqual(math.is_positive_inf(-420), False)
        self.assertEqual(math.is_positive_inf(-69), False)

    def test_is_negative_inf(self):
        self.assertEqual(math.is_negative_inf(float("inf")), False)
        self.assertEqual(math.is_negative_inf(float("-inf")), True)
        self.assertEqual(math.is_negative_inf(float("nan")), False)
        self.assertEqual(math.is_negative_inf(69), False)
        self.assertEqual(math.is_negative_inf(420), False)
        self.assertEqual(math.is_negative_inf(-420), False)
        self.assertEqual(math.is_negative_inf(-69), False)

    def test_delta_angle(self):
        self.assertAlmostEqual(math.delta_angle(69420, 42069), 9)
        self.assertAlmostEqual(math.delta_angle(42069, 69420), -9)
        self.assertAlmostEqual(math.delta_angle(-69420, -42069), -9)
        self.assertAlmostEqual(math.delta_angle(-42069, -69420), 9)
        self.assertAlmostEqual(math.delta_angle(-69420, 42069), -111)
        self.assertAlmostEqual(math.delta_angle(-42069, 69420), -111)

        self.assertAlmostEqual(math.delta_angle(360, 180), 180)
        self.assertAlmostEqual(math.delta_angle(180, 360), 180)

        self.assertAlmostEqual(math.delta_angle(53, 225), 172)
        self.assertAlmostEqual(math.delta_angle(225, 53), -172)

        self.assertAlmostEqual(math.delta_angle(33, 99), 66)
        self.assertAlmostEqual(math.delta_angle(444, 99), 15)

        self.assertAlmostEqual(math.delta_angle(0, 0), 0)
        self.assertAlmostEqual(math.delta_angle(1324, 1324), 0)
        self.assertAlmostEqual(math.delta_angle(-1324, -1324), 0)

        self.assertNotAlmostEqual(math.delta_angle(69, 69), 420)
        self.assertNotAlmostEqual(math.delta_angle(333, 122), 333)
        self.assertNotAlmostEqual(math.delta_angle(122, 122), 122)


class ValueManipulation(unittest.TestCase):
    def test_clamp(self):
        self.assertEqual(math.clamp(0, 0, 1), 0)
        self.assertEqual(math.clamp(-1, 0, 1), 0)
        self.assertEqual(math.clamp(-69, 0, 1), 0)
        self.assertEqual(math.clamp(1, 0, 1), 1)
        self.assertEqual(math.clamp(69, 0, 1), 1)

        self.assertEqual(math.clamp(0.5, 0, 1), 0.5)
        self.assertEqual(math.clamp(-1.22, 0, 1), 0)
        self.assertEqual(math.clamp(3.14159, 0, 1), 1)
        self.assertEqual(math.clamp(0.5555934, 0, 1), 0.5555934)
        self.assertEqual(math.clamp(0.12111, 0, 1), 0.12111)

        self.assertEqual(math.clamp(100000, 42069, 69420), 69420)
        self.assertEqual(math.clamp(100000, 42069, 1000000), 100000)

        self.assertEqual(math.clamp(3, 5, 5), 5)
        self.assertEqual(math.clamp(7, 5, 5), 5)
        self.assertEqual(math.clamp(5, 5, 5), 5)

        self.assertEqual(math.clamp(2, 7, 5), 7)
        self.assertEqual(math.clamp(9, 7, 5), 5)
        self.assertEqual(math.clamp(6, 7, 5), 7)
        self.assertEqual(math.clamp(5, 7, 5), 7)

        self.assertRaises(TypeError, lambda: math.clamp([], {}, "penis")) # type: ignore

    def test_clamp01(self):
        self.assertEqual(math.clamp(0, 0, 1), math.clamp01(0))
        self.assertEqual(math.clamp(-1, 0, 1), math.clamp01(-1))
        self.assertEqual(math.clamp(-69, 0, 1), math.clamp01(-69))
        self.assertEqual(math.clamp(1, 0, 1), math.clamp01(1))
        self.assertEqual(math.clamp(69, 0, 1), math.clamp01(69))
        self.assertEqual(math.clamp(0.5, 0, 1), math.clamp01(0.5))
        self.assertEqual(math.clamp(-1.22, 0, 1), math.clamp01(-1.22))
        self.assertEqual(math.clamp(3.14159, 0, 1), math.clamp01(3.14159))
        self.assertEqual(math.clamp(0.5555934, 0, 1), math.clamp01(0.5555934))
        self.assertEqual(math.clamp(0.12111, 0, 1), math.clamp01(0.12111))

    def test_remap(self):
        self.assertAlmostEqual(math.remap(0.5, 0, 1, 1, 0), 0.5)
        self.assertAlmostEqual(math.remap(0.75, 0, 1, 1, 0), 0.25)

        self.assertAlmostEqual(math.remap(0.5, 0, 1, 0, 2), 1)
        self.assertAlmostEqual(math.remap(0.5, 0, 1, 0, 4), 2)

        self.assertAlmostEqual(math.remap(4, 0, 2, 0, 1), 2)
        self.assertAlmostEqual(math.remap(4, 0, 4, 0, 1), 1)
        self.assertAlmostEqual(math.remap(-4, 0, 4, 0, 1), -1)
        self.assertAlmostEqual(math.remap(-16, 0, 4, 0, 1), -4)
        self.assertAlmostEqual(math.remap(0.25, 0, 4, 0, 1), 0.0625)

    def test_remap01(self):
        self.assertAlmostEqual(math.remap(4, 0, 2, 0, 1), math.remap01(4, 0, 2))
        self.assertAlmostEqual(math.remap(4, 0, 4, 0, 1), math.remap01(4, 0, 4))
        self.assertAlmostEqual(math.remap(-4, 0, 4, 0, 1), math.remap01(-4, 0, 4))
        self.assertAlmostEqual(math.remap(-16, 0, 4, 0, 1), math.remap01(-16, 0, 4))
        self.assertAlmostEqual(math.remap(0.25, 0, 4, 0, 1), math.remap01(0.25, 0, 4))


class Rounding(unittest.TestCase):
    def test_round(self):
        self.assertAlmostEqual(math.round(0.5), 1)
        self.assertAlmostEqual(math.round(1), 1)
        self.assertAlmostEqual(math.round(5), 5)
        self.assertAlmostEqual(math.round(15), 15)
        self.assertAlmostEqual(math.round(15.1), 15)
        self.assertAlmostEqual(math.round(15.25), 15)
        self.assertAlmostEqual(math.round(15.5), 16)
        self.assertAlmostEqual(math.round(15.51), 16)
        self.assertAlmostEqual(math.round(15.75), 16)
        self.assertAlmostEqual(math.round(69.420), 69)

        self.assertAlmostEqual(math.round(-0.5), -1)
        self.assertAlmostEqual(math.round(-1), -1)
        self.assertAlmostEqual(math.round(-5), -5)
        self.assertAlmostEqual(math.round(-15), -15)
        self.assertAlmostEqual(math.round(-15.1), -15)
        self.assertAlmostEqual(math.round(-15.25), -15)
        self.assertAlmostEqual(math.round(-15.5), -16)
        self.assertAlmostEqual(math.round(-15.51), -16)
        self.assertAlmostEqual(math.round(-15.75), -16)
        self.assertAlmostEqual(math.round(-69.420), -69)

    def test_round_to_nearest_n(self):
        self.assertAlmostEqual(math.round_to_nearest_n(0, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(0.5, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(1, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(2, 3), 3)
        self.assertAlmostEqual(math.round_to_nearest_n(5, 3), 6)
        self.assertAlmostEqual(math.round_to_nearest_n(15, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(15.1, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(15.25, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(15.5, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(15.51, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(15.75, 3), 15)
        self.assertAlmostEqual(math.round_to_nearest_n(16.5, 3), 18)
        self.assertAlmostEqual(math.round_to_nearest_n(69.420, 3), 69)

        self.assertAlmostEqual(math.round_to_nearest_n(14.9999, 2), 14)
        self.assertAlmostEqual(math.round_to_nearest_n(15, 2), 16)
        self.assertAlmostEqual(math.round_to_nearest_n(15.25, 2), 16)
        self.assertAlmostEqual(math.round_to_nearest_n(15.5, 2), 16)


        self.assertAlmostEqual(math.round_to_nearest_n(-0, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(-0.5, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(-1, 3), 0)
        self.assertAlmostEqual(math.round_to_nearest_n(-2, 3), -3)
        self.assertAlmostEqual(math.round_to_nearest_n(-5, 3), -6)
        self.assertAlmostEqual(math.round_to_nearest_n(-15, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.1, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.25, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.5, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.51, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.75, 3), -15)
        self.assertAlmostEqual(math.round_to_nearest_n(-16.5, 3), -18)
        self.assertAlmostEqual(math.round_to_nearest_n(-69.420, 3), -69)

        self.assertAlmostEqual(math.round_to_nearest_n(-14.9999, 2), -14)
        self.assertAlmostEqual(math.round_to_nearest_n(-15, 2), -16)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.25, 2), -16)
        self.assertAlmostEqual(math.round_to_nearest_n(-15.5, 2), -16)

    def test_floor_nearest_n(self):
        self.assertAlmostEqual(math.floor_to_nearest_n(0, 3), 0)
        self.assertAlmostEqual(math.floor_to_nearest_n(1, 3), 0)
        self.assertAlmostEqual(math.floor_to_nearest_n(2, 3), 0)
        self.assertAlmostEqual(math.floor_to_nearest_n(3, 3), 3)
        self.assertAlmostEqual(math.floor_to_nearest_n(4, 3), 3)
        self.assertAlmostEqual(math.floor_to_nearest_n(5, 3), 3)
        self.assertAlmostEqual(math.floor_to_nearest_n(6, 3), 6)

        self.assertAlmostEqual(math.floor_to_nearest_n(5.99, 3), 3)
        self.assertAlmostEqual(math.floor_to_nearest_n(6.001, 3), 6)

        self.assertAlmostEqual(math.floor_to_nearest_n(-0, 3), 0)
        self.assertAlmostEqual(math.floor_to_nearest_n(-1, 3), -3)
        self.assertAlmostEqual(math.floor_to_nearest_n(-2, 3), -3)
        self.assertAlmostEqual(math.floor_to_nearest_n(-3, 3), -3)
        self.assertAlmostEqual(math.floor_to_nearest_n(-4, 3), -6)
        self.assertAlmostEqual(math.floor_to_nearest_n(-5, 3), -6)
        self.assertAlmostEqual(math.floor_to_nearest_n(-6, 3), -6)

    def test_ceil_nearest_n(self):
        self.assertAlmostEqual(math.ceil_to_nearest_n(0, 3), 0)
        self.assertAlmostEqual(math.ceil_to_nearest_n(1, 3), 3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(2, 3), 3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(3, 3), 3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(4, 3), 6)
        self.assertAlmostEqual(math.ceil_to_nearest_n(5, 3), 6)
        self.assertAlmostEqual(math.ceil_to_nearest_n(5, 3), 6)
        self.assertAlmostEqual(math.ceil_to_nearest_n(6, 3), 6)

        self.assertAlmostEqual(math.ceil_to_nearest_n(5.99, 3), 6)
        self.assertAlmostEqual(math.ceil_to_nearest_n(6.001, 3), 9)

        self.assertAlmostEqual(math.ceil_to_nearest_n(-0, 3), 0)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-1, 3), 0)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-2, 3), 0)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-3, 3), -3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-4, 3), -3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-5, 3), -3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-6, 3), -6)

        self.assertAlmostEqual(math.ceil_to_nearest_n(-5.99, 3), -3)
        self.assertAlmostEqual(math.ceil_to_nearest_n(-6.001, 3), -6)


class Interpolation(unittest.TestCase):
    def test_lerp(self):
        self.assertAlmostEqual(math.lerp(0, 1, 0.5), 0.5)
        self.assertAlmostEqual(math.lerp(4, 8, 0.5), 6)

        self.assertAlmostEqual(math.lerp(69, 420, 2), 420)
        self.assertAlmostEqual(math.lerp(69, 420, -69), 69)

        self.assertAlmostEqual(math.lerp(0, 4, 0.25), 1)
        self.assertAlmostEqual(math.lerp(2, 8, 0.75), 6.5)

    def test_lerp_unclamped(self):
        self.assertAlmostEqual(math.lerp_unclamped(0, 1, 0.5), 0.5)
        self.assertAlmostEqual(math.lerp_unclamped(4, 8, 0.5), 6)

        self.assertAlmostEqual(math.lerp_unclamped(4, 8, 2), 12)
        self.assertAlmostEqual(math.lerp_unclamped(6, 10, -1), 2)

        self.assertAlmostEqual(math.lerp_unclamped(0, 4, 0.25), 1)
        self.assertAlmostEqual(math.lerp_unclamped(2, 8, 0.75), 6.5)

    def test_lerp_angle(self):
        self.assertAlmostEqual(math.lerp_angle(0, 360, 0.5), 0)
        self.assertAlmostEqual(math.lerp_angle(0, 720, 0.5), 0)

        self.assertAlmostEqual(math.lerp_angle(90, 180, 0.5), 135)
        self.assertAlmostEqual(math.lerp_angle(-45, -90, 0.5), -67.5)

        self.assertAlmostEqual(math.lerp_angle(-69, 420, 2), 60)
        self.assertAlmostEqual(math.lerp_angle(-69, 420, -2), -69)

        self.assertAlmostEqual(math.lerp_angle(180, 360, 0.5), 270)
        self.assertAlmostEqual(math.lerp_angle(360, 720, 0.5), 360)


class Arithmetic(unittest.TestCase):
    def test_kahan_sum(self):
        self.assertEqual(math.kahan_sum(0.1 for _ in range(10)), 1.0) # Iterable
        self.assertEqual(math.kahan_sum([0.1 for _ in range(10)]), 1.0) # List
        self.assertEqual(math.kahan_sum(tuple(0.1 for _ in range(10))), 1.0) # Tuple
        self.assertEqual(math.kahan_sum({0.1 * i: None for i in range(10)}), 4.5) # Dict keys
        self.assertEqual(math.kahan_sum({0.1 * i: None for i in range(10)}.keys()), 4.5) # Dict keys 2
        self.assertEqual(math.kahan_sum({i: 0.1 for i in range(10)}.values()), 1.0) # Dict values

        self.assertEqual(math.kahan_sum(0.1 if i % 2 == 0 else 1 for i in range(10)), 5.5) # int and float mixed
        self.assertEqual(math.kahan_sum(1 for _ in range(10)), 10) # int only

        self.assertRaises(TypeError, lambda: math.kahan_sum(0.1)) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum({i: 0.1 * i for i in range(10)}.items())) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum(None)) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum(object())) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum(...)) # type: ignore

        self.assertRaises(TypeError, lambda: math.kahan_sum([0.1, 0.2, 0.3, 0.4, None])) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum([0.1, 0.2, 0.3, 0.4, object(), 0.6, 0.7])) # type: ignore
        self.assertRaises(TypeError, lambda: math.kahan_sum(None for _ in range(10))) # type: ignore

if __name__ == '__main__':
    unittest.main()



#cython: language_level=3

from libc.math cimport hypot, acos, fabs, INFINITY

from .mvector2 cimport MVector2

cdef extern from "Python.h":
    int SIZEOF_PY_HASH_T

# Hashing
ctypedef unsigned long long int _Vec_uhash_t

cdef _Vec_uhash_t _VecHASH_XXPRIME_1 = 11400714785074694791ULL
cdef _Vec_uhash_t _VecHASH_XXPRIME_2 = 14029467366897019727ULL
cdef _Vec_uhash_t _VecHASH_XXPRIME_5 = 2870177450012600261ULL
cdef inline _Vec_uhash_t _VecHASH_XXROTATE(_Vec_uhash_t x):
    return ((x << 31) | (x >> 33)) # Rotate left 31 bits
# Hashing

cdef double _rad2deg = 180.0 / 3.14159265358979323846

cdef inline double _Vector2Angle(Vector2 _from, Vector2 _to):
    cdef double denom = hypot(_from.x, _from.y) * hypot(_to.x, _to.y) # Multiply magnitudes
    if denom == 0.0:
        return 0.0

    cdef double dot = (_from.x * _to.x) + (_from.y * _to.y)

    cdef double cos_val = dot / denom
    if cos_val < -1.0: # Clamp to range [-1, 1]
        cos_val = -1.0
    elif cos_val > 1.0:
        cos_val = 1.0

    return acos(cos_val) * _rad2deg

cdef class Vector2:
    zero = Vector2(0.0, 0.0)
    one = Vector2(1.0, 1.0)

    up = Vector2(0.0, 1.0)
    down = Vector2(0.0, -1.0)
    left = Vector2(-1.0, 0.0)
    right = Vector2(1.0, 0.0)


    def __init__(self, double x, double y):
        # Vector2(x, y)
        self.x = x
        self.y = y

    def __getitem__(self, Py_ssize_t i):
        # self[i]
        if i == 0:
            return self.x
        if i == 1:
            return self.y

        raise IndexError(i)

    def __repr__(self):
        # repr(self)
        return f"Vector2({self.x}, {self.y})"

    def __len__(self):
        # len(self)
        return 2

    def __add__(self, Vector2 other):
        # self + other
        return Vector2(self.x + other.x, self.y + other.y)

    def __sub__(self, Vector2 other):
        # self - other
        return Vector2(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        # self * other
        cdef double x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x * x, self.y * y)

    def __rmul__(self, other):
        # other * self
        # Duplicated code because inline function didn't work for some reason...
        cdef double x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x * x, self.y * y)

    def __truediv__(self, other):
        # self / other
        cdef double x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x / x, self.y / y)

    def __floordiv__(self, other):
        # self // other
        cdef double x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x // x, self.y // y) # Cython compiles floordiv into C

    def __mod__(self, other):
        # self % other
        cdef double x # Extracted into cdef so that Cython uses C math instead of Python math
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        return Vector2(self.x % x, self.y % y) # Cython compiles modulo into C

    def __divmod__(self, other):
        # divmod(self, other)
        cdef double x
        cdef double y
        if isinstance(other, Vector2):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = other
            y = other
        else:
            return NotImplemented

        cdef double x_fdiv = self.x // x # Cython doesn't have a divmod variant for C
        cdef double x_mod = self.x % x   # We calculate these separately to avoid using Python's divmod
        cdef double y_fdiv = self.y // y
        cdef double y_mod = self.y % y
        return (Vector2(x_fdiv, y_fdiv), Vector2(x_mod, y_mod))

    def __neg__(self):
        # -self
        return Vector2(-self.x, -self.y)

    def __abs__(self):
        # abs(self)
        return Vector2(fabs(self.x), fabs(self.y))

    def __eq__(self, Vector2 other):
        # self == other
        return self.x == other.x and self.y == other.y

    def __iter__(self):
        # iter(self)
        return iter((self.x, self.y))
        # Apparently returning a tuple iterator is faster than writing a custom iterator or yielding...

    # Adapted from tuplehash https://github.com/python/cpython/blob/3.11/Objects/tupleobject.c#L321
    # Doesn't work when extracted into a .pxd file for some reason...
    def __hash__(self):
        if SIZEOF_PY_HASH_T != 8:
            raise RuntimeError("64 bit hash type required.")

        cdef _Vec_uhash_t xlane = <_Vec_uhash_t>hash(self.x)
        cdef _Vec_uhash_t ylane = <_Vec_uhash_t>hash(self.y)

        if xlane == <_Vec_uhash_t>-1 or ylane == <_Vec_uhash_t>-1:
            return -1

        cdef _Vec_uhash_t acc = _VecHASH_XXPRIME_5

        # X
        acc += xlane * _VecHASH_XXPRIME_2
        acc = _VecHASH_XXROTATE(acc)
        acc *= _VecHASH_XXPRIME_1

        # Y
        acc += ylane * _VecHASH_XXPRIME_2
        acc = _VecHASH_XXROTATE(acc)
        acc *= _VecHASH_XXPRIME_1

        acc += (<Py_ssize_t>2) ^ (_VecHASH_XXPRIME_5 ^ 3527539UL)
        # To keep compatibility with tuple's hash implementation
        # The performance improvement by removing this is negligible

        if acc == <_Vec_uhash_t>-1:
            return 1546275796

        return acc

    @property
    def magnitude(self):
        return hypot(self.x, self.y)

    @property
    def normalized(self):
        cdef double magnitude = hypot(self.x, self.y)
        if magnitude == 0.0:
            return Vector2.zero
        return Vector2(self.x / magnitude, self.y / magnitude)

    @staticmethod
    def dot(Vector2 a, Vector2 b):
        return (a.x * b.x) + (a.y * b.y)

    @staticmethod
    def lerp(Vector2 a, Vector2 b, double t):
        if t < 0.0: # clamp01 substitute
            t = 0.0
        elif t > 1.0:
            t = 1.0

        lerp_x = a.x + ((b.x - a.x) * t)
        lerp_y = a.y + ((b.y - a.y) * t)
        return Vector2(lerp_x, lerp_y)

    @staticmethod
    def lerp_unclamped(Vector2 a, Vector2 b, double t):
        lerp_x = a.x + ((b.x - a.x) * t)
        lerp_y = a.y + ((b.y - a.y) * t)
        return Vector2(lerp_x, lerp_y)

    @staticmethod
    def move_towards(Vector2 current, Vector2 target, double max_distance_delta):
        cdef double to_vector_x = target.x - current.x
        cdef double to_vector_y = target.y - current.y

        cdef double dist = hypot(to_vector_x, to_vector_y)

        if dist == 0 or (max_distance_delta >= 0 and dist <= max_distance_delta):
            return target

        cdef double move_x = current.x + to_vector_x / dist * max_distance_delta
        cdef double move_y = current.y + to_vector_y / dist * max_distance_delta
        return Vector2(move_x, move_y)

    @staticmethod
    def smooth_damp(Vector2 current, Vector2 target, MVector2 current_velocity, double smooth_time, double delta_time, double max_speed = <double>INFINITY):
        # Heavy inspiration from https://github.com/Unity-Technologies/UnityCsReference/blob/2023.2/Runtime/Export/Math/Vector2.cs#L289

        smooth_time = max(0.0001, smooth_time)
        cdef double omega = 2.0 / smooth_time

        cdef float x = omega * delta_time
        cdef double exp = 1.0 / (1.0 + x + 0.48 * x * x + 0.235 * x * x * x)

        cdef double change_x = current.x - target.x
        cdef double change_y = current.y - target.y
        cdef Vector2 original_to = target

        cdef double max_change = max_speed * smooth_time
        cdef double dist = hypot(change_x, change_y)
        if dist > max_change:
            change_x = change_x / dist * max_change
            change_y = change_y / dist * max_change

        target = Vector2(current.x - change_x, current.y - change_y)

        cdef double temp_x = (current_velocity.x + omega * change_x) * delta_time
        cdef double temp_y = (current_velocity.y + omega * change_y) * delta_time

        current_velocity.x = (current_velocity.x - omega * temp_x) * exp
        current_velocity.y = (current_velocity.y - omega * temp_y) * exp

        cdef double output_x = target.x + (change_x + temp_x) * exp
        cdef double output_y = target.y + (change_y + temp_y) * exp

        cdef double orig_minus_current_x = original_to.x - current.x
        cdef double orig_minus_current_y = original_to.y - current.y
        cdef double out_minus_orig_x = output_x - original_to.x
        cdef double out_minus_orig_y = output_y - original_to.y

        if (orig_minus_current_x * out_minus_orig_x + orig_minus_current_y * out_minus_orig_y) > 0:
            output_x = original_to.x
            output_y = original_to.y

            current_velocity.x = (output_x - original_to.x) / delta_time
            current_velocity.y = (output_y - original_to.y) / delta_time

        return Vector2(output_x, output_y)


    @staticmethod
    def perpendicular(Vector2 vector):
        return Vector2(-vector.y, vector.x)

    @staticmethod
    def reflect(Vector2 vector, Vector2 normal):
        cdef double dot = (normal.x * vector.x) + (normal.y * vector.y)
        cdef double factor = -2.0 * dot
        return Vector2(factor * normal.x + vector.x, factor * normal.y + vector.y)

    @staticmethod
    def angle(Vector2 _from, Vector2 _to):
        return _Vector2Angle(_from, _to)
        # Extracted into a separate inline method to increase signed_angle performance.
        # From my testing this makes Vector2.angle slower by around 2 % which is basically randomness
        # But on the other hand this change makes Vector2.signed_angle around 12 % faster.

    @staticmethod
    def signed_angle(Vector2 _from, Vector2 _to):
        cdef double unsigned_angle = _Vector2Angle(_from, _to)
        if ((_from.x * _to.y) - (_from.y * _to.x)) < 0.0:
            return -unsigned_angle
        else:
            return unsigned_angle

    @staticmethod
    def distance(Vector2 a, Vector2 b):
        cdef double diff_x = a.x - b.x
        cdef double diff_y = a.y - b.y
        return hypot(diff_x, diff_y)

    @staticmethod
    def min(Vector2 a, Vector2 b):
        """Returns a vector that is made from the smallest components of two vectors."""
        return Vector2(min(a.x, b.x), min(a.y, b.y))

    @staticmethod
    def max(Vector2 a, Vector2 b):
        """Returns a vector that is made from the largest components of two vectors."""
        return Vector2(max(a.x, b.x), max(a.y, b.y))

    def to_tuple(self):
        return (self.x, self.y)

    def to_dict(self):
        return {"x": self.x, "y": self.y}



from typing import ClassVar, Iterator, SupportsIndex
from .mvector2 import MVector2

class Vector2:
    """An immutable two-dimensional vector"""

    zero: ClassVar[Vector2]
    """Shorthand for ``Vector2(0.0, 0.0)``"""
    one: ClassVar[Vector2]
    """Shorthand for ``Vector2(1.0, 1.0)``"""
    up: ClassVar[Vector2]
    """A unit vector pointing up (vector j). Shorthand for ``Vector2(0.0, 1.0)``"""
    down: ClassVar[Vector2]
    """A unit vector pointing down. Shorthand for ``Vector2(0.0, -1.0)``"""
    left: ClassVar[Vector2]
    """A unit vector pointing left. Shorthand for ``Vector2(-1.0, 0.0)``"""
    right: ClassVar[Vector2]
    """A unit vector pointing right (vector i). Shorthand for ``Vector2(1.0, 0.0)``"""

    def __init__(self, x: float, y: float) -> None: ...

    def __getitem__(self, i: SupportsIndex) -> float: ...

    @property
    def x(self) -> float: ...

    @property
    def y(self) -> float: ...

    def __repr__(self) -> str: ...

    def __len__(self) -> int: ...

    def __add__(self, other: Vector2) -> Vector2: ...

    def __sub__(self, other: Vector2) -> Vector2: ...

    def __mul__(self, other: Vector2 | float | int) -> Vector2: ...

    def __rmul__(self, other: Vector2 | float | int) -> Vector2: ...

    def __truediv__(self, other: Vector2 | float | int) -> Vector2: ...

    def __floordiv__(self, other: Vector2 | float | int) -> Vector2: ...

    def __mod__(self, other: Vector2 | float | int) -> Vector2: ...

    def __divmod__(self, other: Vector2 | float | int) -> tuple[Vector2, Vector2]: ...

    def __neg__(self) -> Vector2: ...

    def __abs__(self) -> Vector2: ...

    def __eq__(self, other: Vector2) -> bool: ...

    def __iter__(self) -> Iterator[float]: ...

    def __hash__(self) -> int: ...

    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        ...

    @property
    def normalized(self) -> Vector2:
        """A copy of this vector with a magnitude of 1"""
        ...

    @staticmethod
    def dot(a: Vector2, b: Vector2) -> float:
        """Dot Product of two `Vector2` vectors."""
        ...

    @staticmethod
    def lerp(a: Vector2, b: Vector2, t: float) -> Vector2:
        """
        Linearly interpolates between vectors ``a`` and ``b`` by ``t``.

        The parameter ``t`` is clamped to the range [0, 1].
        """
        ...

    @staticmethod
    def lerp_unclamped(a: Vector2, b: Vector2, t: float) -> Vector2:
        """
        Linearly interpolates between vectors ``a`` and ``b`` by ``t``.

        The parameter ``t`` is not clamped.
        """
        ...

    @staticmethod
    def move_towards(current: Vector2, target: Vector2, max_distance_delta: float) -> Vector2:
        """Moves a point current towards target."""
        ...

    @staticmethod
    def smooth_damp(current: Vector2, target: Vector2, current_velocity: MVector2, smooth_time: float, delta_time: float, max_speed: float = float("inf")) -> Vector2:
        """Gradually changes a vector towards a desired goal over time."""
        ...

    @staticmethod
    def perpendicular(vector: Vector2) -> Vector2:
        """
        Returns a 2D vector with the same magnitude, but perpendicular to the given 2D vector.
        The result is always rotated 90-degrees in a counter-clockwise direction for a 2D coordinate system where the positive Y axis goes up.
        """
        ...

    @staticmethod
    def reflect(vector: Vector2, normal: Vector2) -> Vector2:
        """Reflects a vector off the vector defined by a normal."""
        ...

    @staticmethod
    def angle(_from: Vector2, _to: Vector2) -> float:
        """
        Gets the unsigned angle in degrees between from and to.

        NOTE: The angle returned will always be between 0 and 180 degrees, because the method returns the smallest angle between the vectors.
        """
        ...

    @staticmethod
    def signed_angle(_from: Vector2, _to: Vector2) -> float:
        """
        Gets the signed angle in degrees between from and to. The angle returned is the signed counterclockwise angle between the two vectors.

        NOTE: The angle returned will always be between -180 and 180 degrees, because the method returns the smallest angle between the vectors.
        """
        ...

    @staticmethod
    def distance(a: Vector2, b: Vector2) -> float:
        """Returns the distance between a and b."""
        ...

    @staticmethod
    def min(a: Vector2, b: Vector2) -> Vector2:
        """Create a new `Vector2` by selecting the smallest components of the two given `Vector2` instances."""
        ...

    @staticmethod
    def max(a: Vector2, b: Vector2) -> Vector2:
        """Create a new `Vector2` by selecting the largest components of the two given `Vector2` instances."""
        ...

    def to_tuple(self) -> tuple[float, float]: ...
    def to_dict(self) -> dict[str, float]: ...



from typing import NamedTuple, final

from .. import Rect

@final
class RectOffset(NamedTuple):
    """
    Offsets for rectangles, borders, etc.

    Assumes that X increases to the right and Y increases downwards.
    """

    left: float
    right: float
    top: float
    bottom: float

    @property
    def horizontal(self) -> float:
        return self.left + self.right

    @property
    def vertical(self) -> float:
        return self.top + self.bottom

    def add(self, rect: Rect) -> Rect:
        """Add the border offsets to a rect."""

        return Rect.from_min_max(
            rect.xmin + self.left,
            rect.ymin + self.top,
            rect.xmax - self.right,
            rect.ymax - self.bottom
        )

    def remove(self, rect: Rect) -> Rect:
        """Remove the border offsets from a rect."""

        return Rect.from_min_max(
            rect.xmin - self.left,
            rect.ymin - self.top,
            rect.xmax + self.right,
            rect.ymax + self.bottom
        )

    def to_int_tuple(self) -> tuple[int, int, int, int]:
        left: int = round(self.left)
        right: int = round(self.right)
        top: int = round(self.top)
        bottom: int = round(self.bottom)
        return (left, right, top, bottom)

    def to_float_dict(self) -> dict[str, float]:
        return self._asdict()



from glob import glob
from subprocess import call

CYAN = "\033[36m"
GREEN = "\033[32m"
RED = "\033[31m"
RESET = "\033[0m"

cython_path: str = "cython"
pip_path: str = "pip"

cython_files: list[str] = glob("**/*.pyx", recursive=True)
print(f"Found {len(cython_files)} Cython files.")
for file in cython_files:
    print(CYAN + f"Compiling file '{file}'..." + RESET)
    retcode = call(f"{cython_path} \"{file}\"")
    if retcode == 0:
        print(GREEN + "Compilation successful." + RESET)
    else:
        print(RED + "Compilation failed." + RESET)
        exit(retcode)

print(CYAN + "Installing package..." + RESET)
retcode = call(f"{pip_path} install .")
if retcode == 0:
    print(GREEN + "Install successful." + RESET)
else:
    print(RED + "Install failed." + RESET)
    exit(retcode)



from timeit import timeit
from typing import NamedTuple
from nalpy.math import Vector2, Vector2Int, MVector2, MVector2Int
from nalpy.math import _Legacy_Vector2, _Legacy_Vector2Int, _Legacy_MVector2, _Legacy_MVector2Int # pyright: ignore [reportPrivateUsage]
from nalpy import math
from nalpy.console_utils import set_foreground_color, set_background_color, reset_attributes, ConsoleColor, set_style, ConsoleStyle

setup: str = "x=Vector2(69, 420); y=Vector2(-4, -5); _mutable=Vector2(69, 69); _variant2 = __Variant2(69, 420) if __Variant2 is not None else ..."
commands: tuple[str, ...] = (
#     "Vector2(0, 0)",
#     "x[0]",
#     "x[1]",
#     "x[index_obj]",
#     "x.x",
#     "x.y",
#     "x == x",
#     "x == y",
#     "x.magnitude",

    # Vector specific
#     "Vector2.zero",
#     "x + y",
#     "x - y",
#     "x * y",
#     "x * 2",
#     "2 * x",
#     "x / y",
#     "x / 2",
#     "x // y",
#     "x // 2",
#     "x % y",
#     "x % 2",
#     "divmod(x, y)",
#     "divmod(x, 2)",
#     "-x",
#     "abs(x)",
#     "abs(y)",
#     "hash(x)",
    "for _ in x: pass",

    # MVector specific
#     "Vector2.zero()",
#     "x.copy()",
#     "_mutable.x += 5", # Mutable is reset before the next command, but will be mutated during the different runs of this command.
#     "_mutable.x -= 5",
#     "_mutable += y",
#     "_mutable -= y",
#     "_mutable *= 2",
#     "_mutable //= 2",
#     "_mutable //= 2",
#     "Vector2.from_immutable(_variant2)",
#     "x.to_immutable()"

    # MVector2 Specific
#     "x.normalize()",
#     "x.normalized"

    # Vector2 Specific
#     "x.normalized",
#     "Vector2.dot(x, y)",
#     "Vector2.lerp(x, y, 0.5)",
#     "Vector2.lerp_unclamped(x, y, 1.5)",
#     "Vector2.move_towards(y, x, 15.5)",
#     "Vector2.perpendicular(x)",
#     "Vector2.reflect(x, y)",
#     "Vector2.angle(x, y)",
#     "Vector2.signed_angle(x, y)",
#     "Vector2.distance(x, y)",

    # Vector2Int Specific
#     "Vector2.min(x, y)",
#     "Vector2.max(x, y)",
#     "Vector2.ceil(_math.Vector2(7.5, 6.23))",
#     "Vector2.floor(_math.Vector2(7.5, 6.23))",
#     "Vector2.trunc(_math.Vector2(7.5, 6.23))",
#     "Vector2.round(_math.Vector2(7.5, 6.23))"
)

class Configuration(NamedTuple):
    name: str
    vector2Implementation: type
    variant2: type | None = None

configurations: tuple[Configuration, ...] = (
#     Configuration("Old Vector2", _Legacy_Vector2),
    Configuration("New Vector2", Vector2),
#     Configuration("Old Vector2Int", _Legacy_Vector2Int),
#     Configuration("New Vector2Int", Vector2Int),
#     Configuration("Old MVector2", _Legacy_MVector2, _Legacy_Vector2),
#     Configuration("New MVector2", MVector2, Vector2),
#     Configuration("Old MVector2Int", _Legacy_MVector2Int, _Legacy_Vector2Int),
#     Configuration("New MVector2Int", MVector2Int, Vector2Int)
)

n = 500_000
runs = 50

DELIM: str = 75 * "-"

class _IndexObj:
    def __index__(self) -> int:
        return 0

def run_cmd(cmd: str, config: Configuration, index: int, total: int) -> float:
    gbls = {
        "Vector2": config.vector2Implementation,
        "_math": math,
        "__Variant2": config.variant2,
        "index_obj": _IndexObj()
    }
    secs = timeit(cmd, setup=setup, globals=gbls, number=n)

    print(f"({index + 1} / {total}) {config.name}: {secs} seconds for {n} repetitions")
    return secs

class CommandRan(NamedTuple):
    command: str
    config_run_seconds: tuple[tuple[Configuration, float], ...]

    def get_percent_faster(self) -> tuple[Configuration, Configuration, float]:
        """Returns the percentage of which the first configuration is faster than the last."""
        first_config, first_time = self.config_run_seconds[0]
        last_config, last_time = self.config_run_seconds[-1]
        percent_faster: float = (1 - (last_time / first_time)) * 100
        return (first_config, last_config, percent_faster)

cmds_ran: list[CommandRan] = []

for i, cmd in enumerate(commands):
    print(f"Command: {cmd} ({i + 1} / {len(commands)})")
    configruns: list[tuple[Configuration, float]] = []
    for config in configurations:
        runsofthisconfig: list[float] = []
        for i in range(runs):
            timetaken_seconds: float = run_cmd(cmd, config, i, runs)
            runsofthisconfig.append(timetaken_seconds)
        configruns.append((config, min(runsofthisconfig)))

    cmds_ran.append(CommandRan(cmd, tuple(configruns)))
    print(DELIM)

ran_sorted: tuple[CommandRan, ...] = tuple(sorted(cmds_ran, key=lambda r: r.get_percent_faster()))

def value_to_console_color(value: float):
    if value > 0.0:
        return ConsoleColor.GREEN
    if value < 0.0:
        return ConsoleColor.RED
    return ConsoleColor.YELLOW

for run in ran_sorted:
    set_foreground_color(ConsoleColor.CYAN)
    set_style(ConsoleStyle.BOLD)
    print(run.command, end="")
    reset_attributes()
    print()
    for config, seconds in run.config_run_seconds:
        print(f"{config.name}: {seconds} seconds / {n} runs")
    slowest_config, fastest_config, percent_faster = run.get_percent_faster()
    set_foreground_color(value_to_console_color(percent_faster))
    print(f"{percent_faster}% faster ({slowest_config.name} => {fastest_config.name})", end="")
    reset_attributes()
    print("\n" + DELIM)



from typing import NamedTuple, Self, final, Final

from .. import hypot, ceil, floor, trunc, round, Vector2

@final
class Vector2Int(NamedTuple):
    """An immutable two-dimensional vector with integer precision."""
    x: int
    y: int

    #region Class Properties
    @classmethod
    @property
    def zero(cls) -> Self:
        """Shorthand for ``math.Vector2Int(0, 0)``"""
        return _ZERO #  Returning single instance, because Vector2Int is immutable

    @classmethod
    @property
    def one(cls) -> Self:
        """Shorthand for ``math.Vector2Int(1, 1)``"""
        return _ONE #  Returning single instance, because Vector2Int is immutable


    @classmethod
    @property
    def up(cls) -> Self:
        """A unit vector pointing up (vector j). Shorthand for ``math.Vector2Int(0, 1)``"""
        return _UP #  Returning single instance, because Vector2Int is immutable

    @classmethod
    @property
    def down(cls) -> Self:
        """A unit vector pointing down. Shorthand for ``math.Vector2Int(0, -1)``"""
        return _DOWN #  Returning single instance, because Vector2Int is immutable

    @classmethod
    @property
    def left(cls) -> Self:
        """A unit vector pointing left. Shorthand for ``math.Vector2Int(-1, 0)``"""
        return _LEFT #  Returning single instance, because Vector2Int is immutable

    @classmethod
    @property
    def right(cls) -> Self:
        """A unit vector pointing right (vector i). Shorthand for ``math.Vector2Int(1, 0)``"""
        return _RIGHT #  Returning single instance, because Vector2Int is immutable
    #endregion

    #region Operators

    # Provides __str__ also
    def __repr__(self) -> str:
        """Unambiguous string representation of the vector."""
        return f"Vector2Int({self.x}, {self.y})"

    def __add__(self, other: Self) -> Self:
        """Add (self + other)"""
        if not isinstance(other, Vector2Int):
            return NotImplemented
        return Vector2Int(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Self) -> Self:
        """Subtract (self - other)"""
        if not isinstance(other, Vector2Int):
            return NotImplemented
        return Vector2Int(self.x - other.x, self.y - other.y)

    def __mul__(self, other: Self | int) -> Self:
        """Multiply (self * other)"""
        x: int
        y: int
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2Int(self.x * x, self.y * y)

    def __rmul__(self, other: Self | int) -> Self:
        """Reverse multiply (other * self)"""
        return self.__mul__(other) # other * self = self * other

    def __truediv__(self, other: Self | int | float) -> Vector2:
        """Divide (self / other)"""
        x: float
        y: float
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int | float):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2(self.x / x, self.y / y)

    def __floordiv__(self, other: Self | int) -> Self:
        """Floor Divide (self // other)"""
        x: int
        y: int
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2Int(self.x // x, self.y // y)


    def __mod__(self, other: Self | int) -> Self:
        """Modulo (self % other)"""
        x: int
        y: int
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        return Vector2Int(self.x % x, self.y % y)

    def __divmod__(self, other: Self | int) -> tuple[Self, Self]:
        """Floor division and modulo (divmod(self, other))"""
        x: int
        y: int
        if isinstance(other, Vector2Int):
            x = other.x
            y = other.y
        elif isinstance(other, int):
            x = other
            y = other
        else:
            return NotImplemented

        x_fdiv, x_mod = divmod(self.x, x)
        y_fdiv, y_mod = divmod(self.y, y)
        return (Vector2Int(x_fdiv, y_fdiv), Vector2Int(x_mod, y_mod))


    def __neg__(self) -> Self:
        """Negate (-self)"""
        return Vector2Int(-self.x, -self.y)

    def __abs__(self) -> Self:
        """Absolute value (abs(self))"""
        return Vector2Int(abs(self.x), abs(self.y))

    # __eq__ and __hash__ are provided by NamedTuple

    #endregion

    #region Instance Properties
    @property
    def magnitude(self) -> float:
        """The length of this vector."""
        return hypot(self.x, self.y)
    #endregion

    #region Relation
    @classmethod
    def distance(cls, a: Self, b: Self):
        """Returns the distance between a and b."""
        diff = a - b
        return diff.magnitude
    #endregion

    #region Constructors
    @classmethod
    def ceil(cls, v: Vector2) -> Self:
        return cls(ceil(v.x), ceil(v.y))

    @classmethod
    def floor(cls, v: Vector2) -> Self:
        return cls(floor(v.x), floor(v.y))

    @classmethod
    def round(cls, v: Vector2) -> Self:
        return cls(round(v.x), round(v.y))

    @classmethod
    def trunc(cls, v: Vector2) -> Self:
        return cls(trunc(v.x), trunc(v.y))

    @classmethod
    def min(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the smallest components of two vectors."""
        return cls(min(a.x, b.x), min(a.y, b.y))

    @classmethod
    def max(cls, a: Self, b: Self) -> Self:
        """Returns a vector that is made from the largest components of two vectors."""
        return cls(max(a.x, b.x), max(a.y, b.y))
    #endregion

    #region Conversions
    def to_vector2(self) -> Vector2:
        return Vector2(float(self.x), float(self.y))

    def to_int_dict(self) -> dict[str, int]:
        return self._asdict()
    #endregion

_ZERO: Final[Vector2Int] = Vector2Int(0, 0)
_ONE: Final[Vector2Int] = Vector2Int(1, 1)

_UP: Final[Vector2Int] = Vector2Int(0, 1)
_DOWN: Final[Vector2Int] = Vector2Int(0, -1)
_LEFT: Final[Vector2Int] = Vector2Int(-1, 0)
_RIGHT: Final[Vector2Int] = Vector2Int(1, 0)



#cython: language_level=3

from libc.math cimport hypot

from .vector2 cimport Vector2

cdef class MVector2:
    @staticmethod
    def zero():
        return MVector2(0.0, 0.0)
    @staticmethod
    def one():
        return MVector2(1.0, 1.0)
    @staticmethod
    def up():
        return MVector2(0.0, 1.0)
    @staticmethod
    def down():
        return MVector2(0.0, -1.0)
    @staticmethod
    def left():
        return MVector2(-1.0, 0.0)
    @staticmethod
    def right():
        return MVector2(1.0, 0.0)

    def __init__(self, double x, double y):
        self.x = x
        self.y = y

    @staticmethod
    def from_immutable(Vector2 immutable):
        return MVector2(immutable.x, immutable.y)

    def __getitem__(self, Py_ssize_t i):
        if i == 0:
            return self.x
        if i == 1:
            return self.y

        raise IndexError(i)

    def __repr__(self):
        return f"MVector2({self.x}, {self.y})"

    def __len__(self):
        return 2

    # Only in-place arithmetic supported. For other arithmetic operations, conversion to Vector2 is required.
    def __iadd__(self, other):
        if not isinstance(other, (MVector2, Vector2)):
            return NotImplemented

        self.x += <double>other.x # Casting to force C addition instead of Python addition
        self.y += <double>other.y

        return self

    def __isub__(self, other):
        if not isinstance(other, (MVector2, Vector2)):
            return NotImplemented

        self.x += <double>other.x
        self.y += <double>other.y

        return self

    def __imul__(self, other): # Template from Vector2
        cdef double x
        cdef double y
        if isinstance(other, (MVector2, Vector2)):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        self.x *= x
        self.y *= y

        return self

    def __itruediv__(self, other):
        cdef double x
        cdef double y
        if isinstance(other, (MVector2, Vector2)):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        self.x /= x
        self.y /= y

        return self

    def __ifloordiv__(self, other):
        cdef double x
        cdef double y
        if isinstance(other, (MVector2, Vector2)):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        self.x //= x
        self.y //= y

        return self

    def __imod__(self, other):
        cdef double x
        cdef double y
        if isinstance(other, (MVector2, Vector2)):
            x = other.x
            y = other.y
        elif isinstance(other, (float, int)):
            x = y = other
        else:
            return NotImplemented

        self.x %= x
        self.y %= y

        return self

    def __eq__(self, MVector2 other):
        return self.x == other.x and self.y == other.y

    # NOTE: Object is mutable, do not implement __hash__


    @property
    def magnitude(self):
        return hypot(self.x, self.y)

    @property
    def normalized(self):
        cdef MVector2 v = self.copy()
        v.normalize()
        return v


    cpdef normalize(self) noexcept:
        cdef double magnitude = hypot(self.x, self.y)
        if magnitude == 0.0:
            self.x = self.y = 0.0
        else:
            self.x /= magnitude
            self.y /= magnitude

    cpdef MVector2 copy(self) noexcept:
        return MVector2(self.x, self.y)


    @staticmethod
    def min(MVector2 a, MVector2 b):
        return MVector2(min(a.x, b.x), min(a.y, b.y))

    @staticmethod
    def max(MVector2 a, MVector2 b):
        return MVector2(max(a.x, b.x), max(a.y, b.y))


    def to_immutable(self):
        return Vector2(self.x, self.y)

    def to_tuple(self):
        return (self.x, self.y)

    def to_dict(self):
        return {"x": self.x, "y": self.y}



import nalpy.console_utils._helpers as _helper

# NOTE: Remember to add to console_utils public imports


def cursor_up(count: int = 1):
    _helper.set_seq(count, "A")

def cursor_down(count: int = 1):
    _helper.set_seq(count, "B")

def cursor_forward(count: int = 1):
    _helper.set_seq(count, "C")

def cursor_back(count: int = 1):
    _helper.set_seq(count, "D")

def cursor_line_down(count: int = 1):
    _helper.set_seq(count, "E")

def cursor_line_up(count: int = 1):
    _helper.set_seq(count, "F")


def cursor_hide():
    _helper.set_csi("?25l")

def cursor_show():
    _helper.set_csi("?25h")


def clear():
    """Clear the entire screen and move to upperleft."""
    _helper.erase_display(2)
    cursor_set_pos(1, 1)

def lclear():
    """Clear from cursor to the beginning of the screen."""
    _helper.erase_display(1)

def rclear():
    """Clear from cursor to the end of screen."""
    _helper.erase_display(0)

def erase():
    """Clear the entire line."""
    _helper.erase_line(2)

def lerase():
    """Clear from cursor to beginning of the line."""
    _helper.erase_line(1)

def rerase():
    """Clear from cursor to the end of the line."""
    _helper.erase_line(0)


def scroll_up(lines: int):
    _helper.set_seq(lines, "S")

def scroll_down(lines: int):
    _helper.set_seq(lines, "T")


def cursor_set_pos(row_index: int, column_index: int):
    """rows and columns are 1-based"""
    _helper.set_csi(f"{row_index};{column_index}f")

def cursor_get_pos() -> tuple[int, int]:
    raise NotImplementedError()



from typing import Iterable, overload


def round(__x: float) -> int:
    """
    Round a number to an integer.
    When a number is halfway between two others, it's rounded toward the nearest number that's away from zero.

    This method ends up being around 40 % faster than the built-in Python round function.
    """
    ...

def remap(value: float, from1: float, to1: float, from2: float, to2: float) -> float:
    """Converts a value to another value within the given arguments. Ends are inclusive."""
    ...

def remap01(value: float, from1: float, to1: float) -> float:
    """Converts a value to another value within 0.0 and 1.0. Ends are inclusive."""
    ...

@overload
def clamp(value: int, _min: int, _max: int) -> int:
    """Clamps the value to the specified range. Both ends are inclusive."""
    ...
@overload
def clamp(value: float, _min: float, _max: float) -> float:
    """Clamps the value to the specified range. Both ends are inclusive."""
    ...

def clamp01(value: float) -> float:
    """Shorthand for `math.clamp(value, 0.0, 1.0)`"""
    ...

def delta_angle(current: float, target: float) -> float:
    """Calculates the shortest difference between two given angles."""
    ...

def sign(__x: float) -> int:
    """
    A number that indicates the sign of ``__x``.

    ```
    -1 | less than zero
     0 | equal to zero
    +1 | greater than zero
    ```
    """
    ...


def lerp(a: float, b: float, t: float) -> float:
    """
    Linearly interpolates between ``a`` and ``b`` by ``t``.

    The parameter ``t`` is clamped to the range [0, 1].
    """
    ...

def lerp_unclamped(a: float, b: float, t: float) -> float:
    """
    Linearly interpolates between ``a`` and ``b`` by ``t``.

    The parameter ``t`` is not clamped.
    """
    ...

def lerp_angle(a: float, b: float, t: float) -> float:
    """
    Same as ``lerp``, but makes sure the values interpolate correctly when they wrap around 360 degrees.

    The parameter t is clamped to the range [0, 1]. Variables a and b are assumed to be in degrees.
    """
    ...

def inverse_lerp(a: float, b: float, value: float) -> float:
    """Calculates the clamped parameter ``t`` in ``lerp`` when output is the given ``value``."""
    ...

def smooth_step(a: float, b: float, t: float) -> float:
    """
    Smoothly interpolates between a and b by t.

    The parameter t is clamped to the range [0, 1].
    """
    ...

def move_towards(current: float, target: float, maxDelta: float) -> float:
    """
    Moves a value current towards target.

    This is essentially the same as Lerp, but instead the function will ensure that the speed never exceeds maxDelta. Negative values of maxDelta pushes the value away from target.
    """
    ...

def move_towards_angle(current: float, target: float, maxDelta: float) -> float:
    """
    Same as MoveTowards but makes sure the values interpolate correctly when they wrap around 360 degrees.

    Variables current and target are assumed to be in degrees. For optimization reasons, negative values of maxDelta are not supported and may cause oscillation. To push current away from a target angle, add 180 to that angle instead.
    """
    ...

def ping_pong(t: float, length: float) -> float:
    """
    Returns a value that will increment and decrement between the value 0 and length.

   ``t`` has to be a self-incrementing value.
    """
    ...

def kahan_sum(float_values: Iterable[float]) -> float:
    """
    Implements the Kahan summation algorithm.

    This algorithm greatly reduces the numerical error of the returned floating point values compared to basic summation.
    """



import datetime as _datetime
from enum import Enum as _Enum
from typing import NamedTuple as _NamedTuple
from nalpy import math as _math
from nalpy.humanizer import inflections as _inflections

class _UnitData(_NamedTuple):
    article: str
    name: str
    seconds_length: float
    special_one_past_string: str | None = None
    special_one_future_string: str | None = None

    def get_suffix(self, future: bool) -> str:
        return "from now" if future else "ago"

    def get_special_one_string(self, future: bool) -> str | None:
        return (self.special_one_past_string, self.special_one_future_string)[int(future)] # False => 0, True => 1

    def format_as_words(self, amount: int, future: bool):
        if amount == 1:
            o_str: str | None = self.get_special_one_string(future)
            if o_str is not None:
                return o_str

            return f"{self.article} {self.name} {self.get_suffix(future)}"

        return f"{amount} {self.name}s {self.get_suffix(future)}"

    def format_as_numbers(self, amount: int):
        return _inflections.quick_quantisize(self.name, amount)

class TimeUnit(_Enum): # NOTE: Must be ordered from smallest to largest
    MILLISECOND = _UnitData("a", "millisecond", 0.001)
    SECOND = _UnitData("a", "second", 1.0)
    MINUTE = _UnitData("a", "minute", 60.0)
    HOUR = _UnitData("an", "hour", 3_600.0)
    DAY = _UnitData("a", "day", 86_400, "yesterday", "tomorrow")
    WEEK = _UnitData("a", "week", 604_800)
    MONTH = _UnitData("a", "month", (365.242199 / 12) * 86_400)
    YEAR = _UnitData("a", "year", 365.242199 * 86_400)


def _get_all_units() -> list[_UnitData]:
    return [tu.value for tu in TimeUnit]

def _pick_optimal_unit(seconds: float, min_unit: TimeUnit, max_unit: TimeUnit) -> _UnitData:
    units: list[_UnitData] = _get_all_units()
    optimal_unit_index: int = units.index(max_unit.value)
    while (seconds / units[optimal_unit_index].seconds_length) < 1.0:
        optimal_unit_index -= 1
        if units[optimal_unit_index] == min_unit.value:
            break
    return units[optimal_unit_index]

def humanize_datetime(datetime_: _datetime.datetime, just_now_seconds: float = 10.0, min_unit: TimeUnit = TimeUnit.MILLISECOND, max_unit: TimeUnit = TimeUnit.YEAR) -> str:
    """Humanize a datetime

    Args:
        datetime (datetime): The datetime to humanize.
        just_now_seconds (float, optional): Return 'just now' if the difference in seconds is less than the specified value. -1 to disable. Defaults to 10.0.
        min_unit (TimeUnit, optional): The minimum unit to use. Defaults to TimeUnit.Millisecond.
        max_unit (TimeUnit, optional): The maximum unit to use. Defaults to TimeUnit.Year.

    Returns:
        str: [description]
    """
    now = _datetime.datetime.now()

    delta: _datetime.timedelta = datetime_ - now
    seconds_raw = delta.total_seconds()
    seconds = abs(seconds_raw)

    if seconds < just_now_seconds: # just_now_seconds -1 evaluates to false due to abs()
        return "just now"

    unit = _pick_optimal_unit(seconds, min_unit, max_unit)
    amount = _math.floor(seconds / unit.seconds_length)
    return unit.format_as_words(amount, seconds_raw >= 0.0)

def humanize_timedelta(_timedelta: _datetime.timedelta, precision: int = 1, min_unit: TimeUnit = TimeUnit.MILLISECOND, max_unit: TimeUnit = TimeUnit.YEAR, collection_separator: str = ", ", include_zeros: bool = False, to_words: bool = False) -> str:
    seconds = abs(_timedelta.total_seconds())
    if seconds == 0 and to_words:
        return "no time"

    all_units = _get_all_units()
    optimal = _pick_optimal_unit(seconds, min_unit, max_unit)
    optimal_index = all_units.index(optimal)
    display_units = all_units[max(optimal_index - precision + 1, all_units.index(min_unit.value)):optimal_index + 1]
    display_units.reverse()

    amount_strings: list[str] = []
    remaining_seconds: float = seconds
    for unit in display_units:
        amount = _math.floor(remaining_seconds / unit.seconds_length)
        if amount != 0 or include_zeros:
            amount_strings.append(unit.format_as_numbers(amount))
            remaining_seconds -= amount * unit.seconds_length

    return collection_separator.join(amount_strings)

# def to_ordinal_words(datetime: _datetime.datetime | _datetime.date):
#     raise NotImplementedError()

# def to_clock_notation(time: _datetime.time):
#     raise NotImplementedError()



import typing as _typing

from nalpy import math as _math
from nalpy import console_utils as _console_utils


class ProgressbarStyle(_typing.NamedTuple):
    width: int = 32
    bar_prefix: str = ' '
    bar_suffix: str = ' '
    empty_fill: str = '∙'
    fill: str = '█'
    color: _console_utils.ConsoleColor | None = None
    hide_cursor: bool = True

    @classmethod
    @property
    def default(cls) -> _typing.Self:
        return ProgressbarStyle()

class Progressbar:
    def __init__(self, style: ProgressbarStyle) -> None:
        self._style: ProgressbarStyle = style
        self._cursor_hidden: bool = False
        self._default_message: str | None = None
        self._started: bool = False

    def __enter__(self) -> _typing.Self:
        self.start()
        return self

    def __exit__(self, *_) -> bool:
        self.stop()
        return False

    def start(self, *, default_message: str | None = None):
        """Hides the cursor if it is requested by style.

        Args:
            default_message (str | None, optional): An optional default message to give. Will be written to the console before this progressbar is rendered if provided. Defaults to None.
        """
        if self._started:
            raise RuntimeError("Started already.")

        if self._style.hide_cursor:
            _console_utils.cursor_hide()
            self._cursor_hidden = True

        self._default_message = default_message
        if self._default_message is not None:
            print(self._default_message, end="")

        self._started = True

    def update(self, message: str | None, progress: float, _min: float = 0.0, _max: float = 1.0) -> None:
        """Update this `ProgressBar` instance.

        Args:
            message (str | None): The message to display, use None to display default message.
            progress (float): The progress where _min <= progress <= _max.
            _min (float, optional): The progress minimum value. Defaults to 0.0.
            _max (float, optional): The progress maximum value. Defaults to 1.0.
        """
        if not self._started:
            raise RuntimeError("Not started.")

        if message is None:
            if self._default_message is None:
                err = ValueError("No message provided.")
                err.add_note("No default message given as replacement.")
                raise err
            message = self._default_message

        t = _math.clamp01(_math.remap01(progress, _min, _max))
        filled_length = int(self._style.width * t)
        empty_length = self._style.width - filled_length

        bar = self._style.fill * filled_length
        empty = self._style.empty_fill * empty_length
        suffix = f"{round(t * 100)}%"

        _console_utils.carriage_return()
        print(message, self._style.bar_prefix, sep="", end="")
        if self._style.color is not None:
            _console_utils.set_foreground_color(self._style.color)
        print(bar, end="")
        _console_utils.reset_attributes()
        print(empty, self._style.bar_suffix, suffix, sep="", end="")

    def stop(self, *, end: str | None = "\n") -> None:
        """Resets the cursor visibility if it was modified.

        Args:
            end (str | None, optional): Prints `end` to console using `print` function's rules. Defaults to "\\n".
        """
        if not self._started:
            raise RuntimeError("Not started.")
        self._started = False

        if self._cursor_hidden:
            _console_utils.cursor_show()
            self._cursor_hidden = False

        print(end=end)



import sys as __sys
import typing as __typing

#region public imports of math library functions

# Operations
from math import pow as pow
from math import sqrt as sqrt
from math import cbrt as cbrt

from math import log as log
from math import log10 as log10
from math import log2 as log2
from math import log1p as log1p

# Geometry
from math import tan as tan
from math import atan as atan
from math import atan2 as atan2
from math import sin as sin
from math import asin as asin
from math import cos as cos
from math import acos as acos

from math import degrees as degrees
from math import radians as radians
from math import hypot as hypot

# Rounding
from math import ceil as ceil
from math import floor as floor

# Value checking
from math import isclose as isclose
from math import isinf as isinf
from math import isnan as isnan
from math import isfinite as isfinite

# Value manipulation
from math import trunc as trunc
from math import modf as modf
from math import nextafter as nextafter

# Functions
from math import gcd as gcd
from math import lcm as lcm

#endregion

#region public imports of c extension functions
from nalpy.math._c_extensions.functions import round as round
from nalpy.math._c_extensions.functions import remap as remap
from nalpy.math._c_extensions.functions import remap01 as remap01
from nalpy.math._c_extensions.functions import clamp as clamp
from nalpy.math._c_extensions.functions import clamp01 as clamp01
from nalpy.math._c_extensions.functions import delta_angle as delta_angle
from nalpy.math._c_extensions.functions import sign as sign

from nalpy.math._c_extensions.functions import lerp as lerp
from nalpy.math._c_extensions.functions import lerp_unclamped as lerp_unclamped
from nalpy.math._c_extensions.functions import lerp_angle as lerp_angle
from nalpy.math._c_extensions.functions import inverse_lerp as inverse_lerp
from nalpy.math._c_extensions.functions import smooth_step as smooth_step
from nalpy.math._c_extensions.functions import move_towards as move_towards
from nalpy.math._c_extensions.functions import move_towards_angle as move_towards_angle
from nalpy.math._c_extensions.functions import ping_pong as ping_pong

from nalpy.math._c_extensions.functions import kahan_sum as kahan_sum
#endregion

# Public component imports at the bottom

#region Constants
PI: __typing.Final[float] = 3.14159265358979323846
"""The mathematical constant `3.14159...`"""

EULER: __typing.Final[float] = 2.7182818284590452354
"""Euler's number. The mathematical constant `2.71828...`"""

EPSILON: __typing.Final[float] = __sys.float_info.epsilon
"""Difference between `1.0` and the least value greater than `1.0` that is representable as a float."""

INFINITY: __typing.Final[float] = float("inf")
"""Same as `float("inf")`"""

NEGATIVE_INFINITY: __typing.Final[float] = float("-inf")
"""Same as `float("-inf")`"""

NAN: __typing.Final[float] = float("nan")
"""
Same as `float('nan')`.

NOTE: NaN values are not equal to anything, including themselves. To check if a float is NaN, use `math.isnan()`.
"""

MAXVALUE: __typing.Final[int] = __sys.maxsize
"""
Maximum size of integer-dependant things.

NOTE: Python integers don't have a maximum value.
"""

MINVALUE: __typing.Final[int] = -MAXVALUE - 1
"""
Minimum size of integer-dependant things.

NOTE: Python integers don't have a minimum value.
"""
#endregion

# Initialize TypeVar
_NumberT = __typing.TypeVar("_NumberT", int, float)

#region Basic math functions
def is_positive_inf(__x: __typing.SupportsFloat) -> bool:
    """Return ``True`` if ``x`` is positive infinity, and ``False`` otherwise."""
    x = float(__x)
    return isinf(x) and x > 0 # faster than f == INFINITY, supposedly

def is_negative_inf(__x: __typing.SupportsFloat) -> bool:
    """Return ``True`` if ``x`` is negative infinity, and ``False`` otherwise."""
    x = float(__x)
    return isinf(x) and x < 0 # faster than f == NEGATIVE_INFINITY, supposedly
#endregion

#region Rounding
#region Nearest n
def round_to_nearest_n(__x: __typing.SupportsFloat, n: int) -> int:
    """
    Round a number to the nearest multiple of ``n``.
    When a number is halfway between two multiples of n, it's rounded toward the nearest number that's away from zero.
    """
    if n == 0: # Prevent division by zero. Any multiple of zero is always zero.
        return 0
    return round(float(__x) / n) * n

def floor_to_nearest_n(__x: __typing.SupportsFloat, n: int) -> int:
    """Floor a number to the nearest multiple of ``n``."""
    if n == 0: # Prevent division by zero. Any multiple of zero is always zero.
        return 0
    return floor(float(__x) / n) * n

def ceil_to_nearest_n(__x: __typing.SupportsFloat, n: int) -> int:
    """Ceil a number to the nearest multiple of ``n``."""
    if n == 0: # Prevent division by zero. Any multiple of zero is always zero.
        return 0
    return ceil(float(__x) / n) * n
#endregion

#region Round to digits
def ceil_to_digits(__x: __typing.SupportsFloat, digits: int = 0) -> float:
    """Return the ceiling of x as a float with specified decimal accuracy."""
    pow10: float = pow(10.0, digits)
    return ceil(float(__x) * pow10) / pow10 # dividing changes output to float

def floor_to_digits(__x: __typing.SupportsFloat, digits: int = 0) -> float:
    """Return the floor of x as a float with specified decimal accuracy."""
    pow10: float = pow(10.0, digits)
    return floor(float(__x) * pow10) / pow10 # dividing changes output to float


def round_to_digits(__x: __typing.SupportsFloat, digits: int = 0) -> float:
    """
    Round a number to a given precision in decimal digits.
    When a number is halfway between two others, it's rounded toward the nearest number that's away from zero.

    ``digits`` may be negative.
    """
    pow10: float = pow(10.0, digits)
    return round(float(__x) * pow10) / pow10 # dividing changes output to float


def round_to_nearest_n_to_digits(__x: __typing.SupportsFloat, n: int, digits: int = 0) -> float:
    """Round a number to the nearest multiple of ``n`` with specified decimal accuracy."""
    pow10: float = pow(10.0, digits)
    return round_to_nearest_n(float(__x) * pow10, n) / pow10 # dividing changes output to float

def floor_to_nearest_n_to_digits(__x: __typing.SupportsFloat, n: int, digits: int = 0) -> float:
    """Floor a number to the nearest multiple of ``n`` with specified decimal accuracy."""
    pow10: float = pow(10.0, digits)
    return floor_to_nearest_n(float(__x) * pow10, n) / pow10 # dividing changes output to float

def ceil_to_nearest_n_to_digits(__x: __typing.SupportsFloat, n: int, digits: int = 0) -> float:
    """Ceil a number to the nearest multiple of ``n`` with specified decimal accuracy."""
    pow10: float = pow(10.0, digits)
    return ceil_to_nearest_n(float(__x) * pow10, n) / pow10 # dividing changes output to float
#endregion
#endregion

#region Iterables
def closest(value: int | float, iterable: __typing.Iterable[_NumberT]) -> _NumberT:
    """Return the value in the iterable that is closest to the given value."""
    return min(iterable, key=lambda k: abs(k - value))

def furthest(value: int | float, iterable: __typing.Iterable[_NumberT]) -> _NumberT:
    """Return the value in the iterable that is furthest from the given value."""
    return max(iterable, key=lambda k: abs(k - value))
#endregion

#region Public imports of components
from nalpy.math._c_extensions.vector2 import Vector2 as Vector2
from nalpy.math._c_extensions.vector2_int import Vector2Int as Vector2Int
from nalpy.math._c_extensions.mvector2 import MVector2 as MVector2
from nalpy.math._c_extensions.mvector2_int import MVector2Int as MVector2Int

from nalpy.math._rect.rect import Rect as Rect
from nalpy.math._rect.rect_int import RectInt as RectInt
from nalpy.math._rect.rect_offset import RectOffset as RectOffset
from nalpy.math._rect.rect_offset_int import RectOffsetInt as RectOffsetInt
#endregion

#region Private imports of legacy components
from nalpy.math._legacy_vector2.vector2 import Vector2 as _Legacy_Vector2
from nalpy.math._legacy_vector2.vector2_int import Vector2Int as _Legacy_Vector2Int
from nalpy.math._legacy_vector2.mvector2 import MVector2 as _Legacy_MVector2
from nalpy.math._legacy_vector2.mvector2_int import MVector2Int as _Legacy_MVector2Int
#endregion
